@@ -13,7 +13,6 @@ use Term::Choose           qw();
 use Term::Choose::LineFold qw( line_fold );
 use Term::Choose::Util     qw( get_term_width );
 use Term::Form::ReadLine   qw();
-
 use App::DBBrowser::Auxil;
 
 
@@ -30,24 +29,16 @@ sub new {
 
 sub __session_history {
     my ( $sf ) = @_;
+    my $ax = App::DBBrowser::Auxil->new( $sf->{i}, $sf->{o}, $sf->{d} );
     # Print history:
-    my $quoted_literal = $sf->{i}{driver} =~ /^(?:mysql|MariaDB)\z/ ?
-          qr/ (?<!') ' (?: [^'\\] | \\' | \\\\ )* ' (?!') /x
-        : qr/ (?<!') ' (?: [^']   | ''         )* ' (?!') /x;
-
-    my $iqc = $sf->{d}{identifier_quote_char};
-    my $quoted_identifier = qr/ $iqc (?: [^$iqc] | $iqc$iqc )+ $iqc /x;
-    $quoted_identifier = qr/ $quoted_identifier (?: \. $quoted_identifier ){0,2} /x;
-
-    my $split_rx = qr/ ( $quoted_identifier | $quoted_literal ) /x;
     my $print_history = [];
 
     for my $stmt ( @{$sf->{d}{table_print_history}} ) {
-        $stmt =~ s/\s+\z//;
-        $stmt = join '', map { ! /^[$iqc']/ and s/\s+/ /g; $_ } split $split_rx, $stmt;
-        if ( $sf->{caller} eq 'cte' ) {
-            # nested ctes not allowed
-            $stmt =~ s/^WITH\s.+\)\s(SELECT\s.+)\z/$1/;
+        $stmt = $ax->normalize_space_in_stmt( $stmt );
+        $stmt =~ s/^ WITH \s .+ \) \s ( SELECT \s .+ ) \z/$1/ix;
+        if ( $stmt =~ /^ SELECT \s \* \s FROM \s \(( SELECT \s .+ ) \) \s \S+ \z/ix ) {
+            # union stmt
+            $stmt = $1;
         }
         if ( any { $_ eq $stmt } @$print_history ) {
             next;
@@ -162,9 +153,8 @@ sub __choose_query {
 }
 
 
-sub prepare_and_add_cte {
+sub __prepare_and_add_cte {
     my ( $sf, $sql ) = @_;
-    $sf->{caller} = 'cte';
     my $ax = App::DBBrowser::Auxil->new( $sf->{i}, $sf->{o}, $sf->{d} );
     my $tc = Term::Choose->new( $sf->{i}{tc_default} );
     my $tr = Term::Form::ReadLine->new( $sf->{i}{tr_default} );
@@ -192,24 +182,28 @@ sub prepare_and_add_cte {
 
             NAME: while ( 1 ) {
                 # Readline
-                my $name = $tr->readline(
+                my $full_cte_name = $tr->readline(
                     'Name : ',
                     { default => $default_name, show_context => 1, info => $info, history => [ 'cte1' .. 'cte9' ] }
                 );
                 $ax->print_sql_info( $info );
-                if ( ! length $name ) {
+                if ( ! length $full_cte_name ) {
                     next CHOOSE_QUERY if ++$count_query_loop > 1;
                     next QUERY;
                 }
-                my $qc = $sf->{d}{identifier_quote_char};
-                my $table = $name =~ s/^ \s* (?:recursive\s)? \s* ( $qc(?:[^$qc]|$qc$qc)+$qc | \S+ ) \s* (?:\(.+)? \z/$1/rix;
-                my $cte = { name => $name, query => $query, table => $table };
-                $default_name = length $default_name ? '' : $table;
+                my $rx_quoted_identifier = $ax->regex_quoted_identifier();
+                my $cte_name = $full_cte_name =~ s/^ \s* (?:RECURSIVE\s+)? ( $rx_quoted_identifier | [^\s(]+ ) \s* (?:\(.+)? \z/$1/rix;
+                my $cte = {
+                    full_name => $full_cte_name,
+                    query => $query,
+                    name => $cte_name,
+                };
+                $default_name = length $default_name ? '' : $full_cte_name;
                 my @table_names = keys %{$sf->{d}{tables_info}};
-                my @cte_names = map { $_->{table} } @{$sql->{ctes}};
-                if ( any { $_ eq $table } @table_names, @cte_names ) {
-                    my $type = ( firstidx { $_ eq $table } @table_names ) > -1 ? 'table' : 'cte';
-                    my $prompt = "A $type '$table' already exists.";
+                my @cte_names = map { $_->{name} } @{$sql->{ctes}};
+                if ( any { $_ eq $cte_name } @table_names, @cte_names ) {
+                    my $type = ( firstidx { $_ eq $cte_name } @table_names ) > -1 ? 'table' : 'cte';
+                    my $prompt = "A $type '$cte_name' already exists.";
                     # Choose
                     $tc->choose(
                         [ 'Press ENTER' ],
@@ -219,9 +213,13 @@ sub prepare_and_add_cte {
                     next NAME;
                 }
                 else {
+
+                    if ( $cte->{full_name} =~ s/^\s*RECURSIVE\s+//i ) {
+                        $cte->{is_recursive} = 1;
+                    }
                     push @{$sql->{ctes}}, $cte;
                     unshift @{$sf->{d}{subquery_history}}, $query;
-                    return $table;
+                    return $cte_name;
                 }
             }
         }
@@ -229,9 +227,114 @@ sub prepare_and_add_cte {
 }
 
 
+sub choose_cte {
+    my ( $sf, $sql ) = @_;
+    my $ax = App::DBBrowser::Auxil->new( $sf->{i}, $sf->{o}, $sf->{d} );
+    my $tc = Term::Choose->new( $sf->{i}{tc_default} );
+    my ( $new_cte, $reset_unused ) = ( '  New CTE', '  Reset' );
+    my @pre = ( undef );
+    my $old_idx = 0;
+
+    CHOOSE_CTE: while ( 1 ) {
+        my @avail_ctes = map { '- ' . $_->{name} } @{$sql->{ctes}};
+        my $menu = [ @pre, @avail_ctes, $new_cte, $reset_unused ];
+        my $info = $ax->get_sql_info( $sql );
+        # Choose
+        my $idx = $tc->choose(
+            $menu,
+            { %{$sf->{i}{lyt_v}}, info => $info, index => 1, undef => '  <=', default => $old_idx }
+        );
+        $ax->print_sql_info( $info );
+        if ( ! defined $idx || $idx == 0 ) {
+            return;
+        }
+        if ( $sf->{o}{G}{menu_memory} ) {
+            if ( $old_idx == $idx && ! $ENV{TC_RESET_AUTO_UP} ) {
+                $old_idx = 0;
+                next CHOOSE_CTE;
+            }
+            $old_idx = $idx;
+        }
+        if ( $menu->[$idx] eq $new_cte ) {
+            $sf->__prepare_and_add_cte( $sql );
+            $sf->{d}{cte_history} = [ @{$sql->{ctes}} ];
+            $old_idx++;
+        }
+        elsif ( $menu->[$idx] eq $reset_unused ) {
+            my $stmt_type = $sf->{d}{stmt_types}[0];
+            my $bu_ctes = [ @{$sql->{ctes}} ];
+            $sql->{ctes} = [];
+            my $stmt = $ax->get_stmt( $sql, $stmt_type, 'prepare' );
+            $sql->{ctes} = $bu_ctes;
+            if ( $stmt_type eq 'Join' ) {
+                $stmt = "SELECT * FROM " . $stmt;
+            }
+            $sf->__reset_unused_ctes( $sql, $stmt );
+            $sf->{d}{cte_history} = [ @{$sql->{ctes}} ];
+        }
+        else {
+            $idx -= @pre;
+            my $cte_name = $sql->{ctes}[$idx]{name};
+            return $cte_name;
+        }
+    }
+}
+
+
+sub __table_names_in_query {
+    my ( $sf, $stmt ) = @_;
+    my $ax = App::DBBrowser::Auxil->new( $sf->{i}, $sf->{o}, $sf->{d} );
+    my $rx_quoted_literal = $ax->regex_quoted_literal();
+    my $rx_quoted_identifier = $ax->regex_quoted_identifier();
+    my $rx_sql_elem = qr/ (?: $rx_quoted_identifier | [^',\s]+ ) /x;
+    $rx_sql_elem = qr/ $rx_sql_elem (?: \. $rx_sql_elem ){0,2} /x;
+    my $rx_split_stmt = qr/ ( $rx_quoted_literal | $rx_sql_elem | \s*,\s* ) \s+ /x;
+    my $tables = [];
+    my $implicit_join;
+    my @token = grep { length } split $rx_split_stmt, $stmt;
+
+    while ( @token ) {
+        my $t = shift @token;
+        if( $t =~ /^JOIN\z/i ) {
+            if ( @token ) {
+                push @$tables, shift @token;
+            }
+        }
+        elsif ( $t =~ /^FROM\z/i || $implicit_join ) {
+            if ( @token ) {
+                push @$tables, shift @token;
+            }
+            if ( defined $token[1] && $token[1] eq ',' ) {
+                shift @token;
+            }
+            elsif ( defined $token[2] && $token[2] eq ',' && $token[0] =~ /^AS\z/i ) {
+                shift @token;
+                shift @token;
+            }
+            $implicit_join = defined $token[0] && $token[0] eq ',' ? 1 : 0;
+        }
+    }
+    return $tables;
+}
+
+
+sub __reset_unused_ctes {
+    my ( $sf, $sql, $stmt ) = @_;
+    my $all_tables_used_in_stmts = $sf->__table_names_in_query( $stmt );
+
+    CTE: for my $cte ( reverse @{$sql->{ctes}} ) {
+        if ( any { $_ eq $cte->{name} } @$all_tables_used_in_stmts ) {
+            my $tables_in_cte = $sf->__table_names_in_query( $cte->{query} );
+            push @$all_tables_used_in_stmts, @$tables_in_cte;
+            $cte->{keep} = 1;
+        }
+    }
+    $sql->{ctes} = [ grep { delete $_->{keep} } @{$sql->{ctes}} ];
+}
+
+
 sub subquery {
     my ( $sf, $sql ) = @_;
-    $sf->{caller} = 'subquery';
     my $ax = App::DBBrowser::Auxil->new( $sf->{i}, $sf->{o}, $sf->{d} );
     my $tr = Term::Form::ReadLine->new( $sf->{i}{tr_default} );
 
